/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>

#include <doca_argp.h>
#include <doca_log.h>

#include "common_common.h"

DOCA_LOG_REGISTER(SYNC_EVENT::MAIN);

/*
 * Sample's logic
 *
 * @se_cfg [in]: user configuration represents command line arguments
 * @se_rt_objs [in/out]: sample's runtime resources
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t sync_event_run(struct sync_event_config *se_cfg, struct sync_event_runtime_objects *se_rt_objs);

/*
 * Sample main function
 *
 * @argc [in]: command line arguments size
 * @argv [in]: array of command line arguments
 * @return: EXIT_SUCCESS on success and EXIT_FAILURE otherwise
 */
int
main(int argc, char **argv)
{
	struct doca_logger_backend *stdout_logger = NULL;
	doca_error_t result = DOCA_SUCCESS;
	struct sync_event_config se_cfg;
	struct sync_event_runtime_objects se_rt_objs;

	memset(&se_cfg, 0, sizeof(struct sync_event_config));
	memset(&se_rt_objs, 0, sizeof(struct sync_event_runtime_objects));

	result = doca_log_create_file_backend(stdout, &stdout_logger);
	if (result != DOCA_SUCCESS)
		return EXIT_FAILURE;

#ifndef DOCA_ARCH_HOST
	DOCA_LOG_ERR("Sample can run only on the Host");
	return EXIT_FAILURE;
#endif

	result = doca_argp_init("doca_sync_event_host", &se_cfg);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init argp: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}

	result = sync_event_params_register();
	if (result != DOCA_SUCCESS) {
		doca_argp_destroy();
		DOCA_LOG_ERR("Failed to register sample parameters: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}

	result = doca_argp_start(argc, argv);
	if (result != DOCA_SUCCESS) {
		doca_argp_destroy();
		DOCA_LOG_ERR("Failed to parse sample input: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}

	result = doca_pci_bdf_from_string(se_cfg.dev_pci_addr, &se_rt_objs.dev_pci_addr);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse device PCI address: %s", doca_get_error_string(result));
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	result = sync_event_run(&se_cfg, &se_rt_objs);
	if (result != DOCA_SUCCESS) {
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	doca_argp_destroy();

	return DOCA_SUCCESS;
}
