/*
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <string.h>

#include <doca_argp.h>
#include <doca_error.h>
#include <doca_dev.h>
#include <doca_log.h>

#include <doca_utils.h>

#include "dma_common.h"

DOCA_LOG_REGISTER(DPU_LOCAL_DMA_COPY::MAIN);

/* Sample's Logic */
doca_error_t dma_local_copy(struct doca_pci_bdf *pcie_addr, char *dst_buffer, const char *src_buffer, size_t length);

/*
 * Sample main function
 *
 * @argc [in]: command line arguments size
 * @argv [in]: array of command line arguments
 * @return: EXIT_SUCCESS on success and EXIT_FAILURE otherwise
 */
int
main(int argc, char **argv)
{
	struct dma_config dma_conf;
	struct doca_pci_bdf pcie_dev;
	char *dst_buffer, *src_buffer;
	size_t length;
	doca_error_t result;
	struct doca_logger_backend *stdout_logger = NULL;

	/* Set the default configuration values (Example values) */
	strcpy(dma_conf.pci_address, "03:00.0");
	strcpy(dma_conf.cpy_txt, "This is a sample piece of text");
	/* No need to set export_desc_path and buf_info_path which are only needed for DMA across devices */

	/* Create a logger backend that prints to the standard output */
	result = doca_log_create_file_backend(stdout, &stdout_logger);
	if (result != DOCA_SUCCESS)
		return EXIT_FAILURE;

	result = doca_argp_init("doca_dma_local_copy", &dma_conf);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init ARGP resources: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}
	result = register_dma_params();
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to register DMA sample parameters: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}
	result = doca_argp_start(argc, argv);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse sample input: %s", doca_get_error_string(result));
		return EXIT_FAILURE;
	}
#ifndef DOCA_ARCH_DPU
	DOCA_LOG_ERR("Local DMA copy can run only on the DPU");
	doca_argp_destroy();
	return EXIT_FAILURE;
#endif
	length = strlen(dma_conf.cpy_txt) + 1;
	dst_buffer = (char *)calloc(1, length);
	if (dst_buffer == NULL) {
		DOCA_LOG_ERR("Destination buffer allocation failed");
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	src_buffer = (char *)malloc(length);
	if (src_buffer == NULL) {
		DOCA_LOG_ERR("Source buffer allocation failed");
		free(dst_buffer);
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	memcpy(src_buffer, dma_conf.cpy_txt, length);

	result = doca_pci_bdf_from_string(dma_conf.pci_address, &pcie_dev);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to parse pci address: %s", doca_get_error_string(result));
		free(src_buffer);
		free(dst_buffer);
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	result = dma_local_copy(&pcie_dev, dst_buffer, src_buffer, length);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Sample function has failed: %s", doca_get_error_string(result));
		free(src_buffer);
		free(dst_buffer);
		doca_argp_destroy();
		return EXIT_FAILURE;
	}

	free(src_buffer);
	free(dst_buffer);
	doca_argp_destroy();

	return EXIT_SUCCESS;
}
