/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <unistd.h>

#include <infiniband/mlx5dv.h>

#include <doca_error.h>
#include <doca_log.h>

#include "dpa_common.h"

DOCA_LOG_REGISTER(KERNEL_LAUNCH::SAMPLE);

/* Kernel function decleration */
extern doca_dpa_func_t hello_world;

/*
 * Create DPA event with the appropriate update and wait flags for completion event: updated by the DPA and waited
 * on by the CPU.
 *
 * @doca_dpa [in]: Previously created DPA context
 * @comp_event [out]: Created event
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
create_dpa_completion_event(struct doca_dpa *doca_dpa, struct doca_dpa_event **comp_event)
{
	doca_error_t result;

	result = doca_dpa_event_create(doca_dpa, DOCA_DPA_EVENT_ACCESS_DPA, DOCA_DPA_EVENT_ACCESS_CPU,
					DOCA_DPA_EVENT_WAIT_DEFAULT, comp_event, 0);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create host completion event: %s", doca_get_error_string(result));
		return result;
	}

	return result;
}

/*
 * Run kernel_launch sample
 *
 * @resources [in]: DOCA DPA resources that the DPA sample will use
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t
kernel_launch(struct dpa_resources *resources)
{
	struct doca_dpa_event *wait_event = NULL;
	struct doca_dpa_event *comp_event = NULL;
	/* Wait event threshold */
	uint64_t wait_thresh = 4;
	/* Completion event val */
	uint64_t comp_event_val = 10;
	/* Number of DPA threads */
	unsigned int n_threads = 1;
	doca_error_t result;
	doca_error_t tmp_result;

	/* Creating DOCA DPA events */
	result = create_dpa_completion_event(resources->doca_dpa, &comp_event);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create DOCA DPA events: %s", doca_get_error_string(result));
		return result;
	}

	DOCA_LOG_INFO("All DPA resources have been created");

	/* kernel launch */
	result = doca_dpa_kernel_launch(resources->doca_dpa, wait_event, wait_thresh, comp_event, comp_event_val, DOCA_DPA_EVENT_OP_SET, &hello_world, n_threads);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to launch hello_world kernel: %s", doca_get_error_string(result));
		goto destroy_event;
	}

	/* Wait until completion event reach completion val */
	result = doca_dpa_event_wait_until(comp_event, comp_event_val, DOCA_DPA_EVENT_CMP_GEQ);
	if (result != DOCA_SUCCESS)
		DOCA_LOG_ERR("Failed to wait for host completion event: %s", doca_get_error_string(result));

destroy_event:
	/* Sleep 2 seconds before destroying DPA events */
	sleep(2);

	/* destroy events */
	tmp_result = doca_dpa_event_destroy(comp_event);
	if (tmp_result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to destroy host completion event: %s", doca_get_error_string(tmp_result));
		DOCA_ERROR_PROPAGATE(result, tmp_result);
	}

	return result;
}
