/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>

#include <infiniband/mlx5dv.h>

#include <doca_error.h>
#include <doca_log.h>

#include "dpa_common.h"

DOCA_LOG_REGISTER(WAIT_KERNEL_LAUNCH::SAMPLE);

/* Kernel function decleration */
extern doca_dpa_func_t hello_world;

/*
 * Waits for 5 seconds and then update a given event.
 * This function is made to be used by a pthread, hence it needs to return void*, and it will always return NULL.
 *
 * @event [in]: event to be updated
 * @return: NULL (dummy return)
 */
static void *
wait_event_update_thread(void *event)
{
	sleep(5);
	doca_dpa_event_update((struct doca_dpa_event *)event, 5, DOCA_DPA_EVENT_OP_SET);

	/* Dummy return */
	return NULL;
}

/*
 * Create two DPA events with the appropriate update and wait flags for wait and completion event:
 *	Wait event is updated by the CPU and waited on by the DPA.
 *	Completion event is updated by the DPA and waited on by the CPU.
 *
 * @doca_dpa [in]: Previously created DPA context
 * @wait_event [out]: Created event
 * @comp_event [out]: Created event
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t
create_dpa_events(struct doca_dpa *doca_dpa, struct doca_dpa_event **wait_event, struct doca_dpa_event **comp_event)
{
	doca_error_t result;

	result = doca_dpa_event_create(doca_dpa, DOCA_DPA_EVENT_ACCESS_CPU, DOCA_DPA_EVENT_ACCESS_DPA,
					DOCA_DPA_EVENT_WAIT_DEFAULT, wait_event, 0);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create host wait event: %s", doca_get_error_string(result));
		return result;
	}

	result = doca_dpa_event_create(doca_dpa, DOCA_DPA_EVENT_ACCESS_DPA, DOCA_DPA_EVENT_ACCESS_CPU,
					DOCA_DPA_EVENT_WAIT_DEFAULT, comp_event, 0);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create host completion event: %s", doca_get_error_string(result));
		doca_dpa_event_destroy(*wait_event);
		return result;
	}

	return result;
}

/*
 * Run wait_kernel_launch sample
 *
 * @resources [in]: DOCA DPA resources that the DPA sample will use
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t
wait_kernel_launch(struct dpa_resources *resources)
{
	struct doca_dpa_event *wait_event = NULL;
	struct doca_dpa_event *comp_event = NULL;
	pthread_t tid = 0;
	/* Wait event threshold */
	uint64_t wait_thresh = 4;
	/* Completion event val */
	uint64_t comp_event_val = 10;
	doca_error_t result;
	doca_error_t tmp_result;
	int res = 0;

	/* Creating DOCA DPA events */
	result = create_dpa_events(resources->doca_dpa, &wait_event, &comp_event);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create DOCA DPA events: %s", doca_get_error_string(result));
		return result;
	}

	DOCA_LOG_INFO("All DPA resources have been created");

	/* kernel launch */
	result = doca_dpa_kernel_launch(resources->doca_dpa, wait_event, wait_thresh, comp_event, comp_event_val, DOCA_DPA_EVENT_OP_SET, &hello_world, 1);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to launch hello_world kernel: %s", doca_get_error_string(result));
		goto destroy_events;
	}

	/* update wait event within a different CPU thread */
	res = pthread_create(&tid, NULL, wait_event_update_thread, (void *)wait_event);
	if (res != 0) {
		DOCA_LOG_ERR("Failed to create thread");
		result = DOCA_ERROR_OPERATING_SYSTEM;
		goto destroy_events;
	}

	res = pthread_detach(tid);
	if (res != 0) {
		DOCA_LOG_ERR("pthread_detach() failed");
		result = DOCA_ERROR_OPERATING_SYSTEM;
		goto destroy_events;
	}

	/* Wait until completion event reach completion val */
	result = doca_dpa_event_wait_until(comp_event, comp_event_val, DOCA_DPA_EVENT_CMP_GEQ);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to wait for host completion event: %s", doca_get_error_string(result));
		goto destroy_events;
	}

destroy_events:
	/* Sleep 2 seconds before destroying DPA events */
	sleep(2);

	/* destroy events */
	tmp_result = doca_dpa_event_destroy(wait_event);
	if (tmp_result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to destroy host wait event: %s", doca_get_error_string(tmp_result));
		DOCA_ERROR_PROPAGATE(result, tmp_result);
	}

	tmp_result = doca_dpa_event_destroy(comp_event);
	if (tmp_result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to destroy host completion event: %s", doca_get_error_string(tmp_result));
		DOCA_ERROR_PROPAGATE(result, tmp_result);
	}

	return result;
}
