/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <rte_byteorder.h>
#include <rte_random.h>

#include <doca_flow.h>
#include <doca_log.h>

#include "flow_common.h"

DOCA_LOG_REGISTER(FLOW_AGING);

/* user data struct for each entry */
struct sample_user_data {
	int entry_num;				/* entry number */
	int port_id;				/* port ID of the entry */
	struct doca_flow_pipe_entry *entry;	/* entry pointer */
};

/*
 * Create DOCA Flow pipe with five tuple match and monitor with aging flag
 *
 * @port [in]: port of the pipe
 * @port_id [in]: port ID of the pipe
 * @pipe [out]: created pipe pointer
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise.
 */
static doca_error_t
create_aging_pipe(struct doca_flow_port *port, int port_id, struct doca_flow_pipe **pipe)
{
	struct doca_flow_match match;
	struct doca_flow_monitor monitor;
	struct doca_flow_actions actions;
	struct doca_flow_actions *actions_arr[NB_ACTIONS_ARR];
	struct doca_flow_fwd fwd;
	struct doca_flow_pipe_cfg pipe_cfg;

	memset(&match, 0, sizeof(match));
	memset(&monitor, 0, sizeof(monitor));
	memset(&actions, 0, sizeof(actions));
	memset(&fwd, 0, sizeof(fwd));
	memset(&pipe_cfg, 0, sizeof(pipe_cfg));

	pipe_cfg.attr.name = "AGING_PIPE";
	pipe_cfg.attr.type = DOCA_FLOW_PIPE_BASIC;
	pipe_cfg.match = &match;
	pipe_cfg.monitor = &monitor;
	actions_arr[0] = &actions;
	pipe_cfg.actions = actions_arr;
	pipe_cfg.attr.is_root = true;
	pipe_cfg.attr.nb_actions = NB_ACTIONS_ARR;
	pipe_cfg.port = port;

	/* set monitor with aging */
	monitor.flags = DOCA_FLOW_MONITOR_AGING;

	/* 5 tuple match */
	match.outer.l4_type_ext = DOCA_FLOW_L4_TYPE_EXT_TCP;
	match.outer.l3_type = DOCA_FLOW_L3_TYPE_IP4;
	match.outer.ip4.src_ip = 0xffffffff;
	match.outer.ip4.dst_ip = 0xffffffff;
	match.outer.tcp.l4_port.src_port = 0xffff;
	match.outer.tcp.l4_port.dst_port = 0xffff;

	fwd.type = DOCA_FLOW_FWD_PORT;
	fwd.port_id = port_id ^ 1;
	return doca_flow_pipe_create(&pipe_cfg, &fwd, NULL, pipe);
}

/*
 * Add DOCA Flow entries to the aging pipe, each entry with different aging time
 *
 * @pipe [in]: pipe of the entries
 * @user_data [in]: user data array
 * @port [in]: port of the entries
 * @port_id [in]: port ID of the entries
 * @num_of_aging_entries [in]: number of entries to add
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise.
 */
static doca_error_t
add_aging_pipe_entries(struct doca_flow_pipe *pipe, struct sample_user_data *user_data,
		struct doca_flow_port *port, int port_id, int num_of_aging_entries)
{
	struct doca_flow_match match;
	struct doca_flow_actions actions;
	struct doca_flow_monitor monitor;
	struct entries_status status;
	int i;
	enum doca_flow_flags_type flags = DOCA_FLOW_WAIT_FOR_BATCH;
	doca_be32_t dst_ip_addr = BE_IPV4_ADDR(8, 8, 8, 8);
	doca_be16_t dst_port = rte_cpu_to_be_16(80);
	doca_be16_t src_port = rte_cpu_to_be_16(1234);
	doca_be32_t src_ip_addr; /* set different src ip per entry */
	doca_error_t result;

	memset(&status, 0, sizeof(status));

	for (i = 0; i < num_of_aging_entries; i++) {
		src_ip_addr = BE_IPV4_ADDR(i, 2, 3, 4);

		memset(&match, 0, sizeof(match));
		memset(&actions, 0, sizeof(actions));
		memset(&monitor, 0, sizeof(monitor));

		monitor.flags |= DOCA_FLOW_MONITOR_AGING;
		/* set user data struct in monitor */
		monitor.user_data = (void *)(&user_data[i]);
		/* flows will be aged out in 5 - 60s */
		monitor.aging = (uint32_t)rte_rand() % 55 + 5;

		match.outer.ip4.dst_ip = dst_ip_addr;
		match.outer.ip4.src_ip = src_ip_addr;
		match.outer.tcp.l4_port.dst_port = dst_port;
		match.outer.tcp.l4_port.src_port = src_port;
		/* fill user data with entry number and entry pointer */
		user_data[i].entry_num = i;
		user_data[i].port_id = port_id;

		if (i == num_of_aging_entries - 1)
			flags = DOCA_FLOW_NO_WAIT; /* send the last entry with DOCA_FLOW_NO_WAIT flag for pushing all the entries */

		result = doca_flow_pipe_add_entry(0, pipe, &match, &actions, &monitor, NULL, flags, &status,
						  &user_data[i].entry);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to add entry: %s", doca_get_error_string(result));
			return result;
		}
	}
	do {
		result = doca_flow_entries_process(port, 0, DEFAULT_TIMEOUT_US, num_of_aging_entries - status.nb_processed);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to process entries: %s", doca_get_error_string(result));
			return result;
		}
		if (status.failure) {
			DOCA_LOG_ERR("Failed to process entries, status is not success");
			return DOCA_ERROR_BAD_STATE;
		}
	} while (status.nb_processed < num_of_aging_entries);

	return DOCA_SUCCESS;
}

/*
 * Handle all aged flow in a port
 *
 * @port [in]: port to remove the aged flow from
 * @aged_entries [in]: allocated array for the aged flows
 * @num_of_aging_entries [in]: number of entries in a port
 * @total_counter [in/out]: counter for all aged flows in both ports
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise.
 */
static doca_error_t
handle_aged_flow(struct doca_flow_port *port, struct doca_flow_aged_query *aged_entries, int num_of_aging_entries, int *total_counter)
{
	uint64_t quota_time = 20; /* max handling aging tile in ms */
	int num_of_aged_entries = 0;
	doca_error_t result;
	enum doca_flow_flags_type flags;
	struct sample_user_data *aged_entry_data;
	struct entries_status status;
	int i;

	memset(&status, 0, sizeof(status));

	num_of_aged_entries = doca_flow_aging_handle(port, 0, quota_time, aged_entries, num_of_aging_entries);
	/* call handle aging until full cycle complete */
	while (num_of_aged_entries != -1) {
		*total_counter += num_of_aged_entries;
		DOCA_LOG_INFO("Num of aged entries: %d, total: %d", num_of_aged_entries, *total_counter);
		flags = DOCA_FLOW_WAIT_FOR_BATCH;
		status.nb_processed = 0;
		for (i = 0; i < num_of_aged_entries; i++) {
			aged_entry_data = (struct sample_user_data *)((void *)aged_entries[i].user_data);
			DOCA_LOG_INFO(
				"Entry number %d from port %d aged out and removed",
				aged_entry_data->entry_num,
				aged_entry_data->port_id);
			if (i == num_of_aged_entries - 1)
				flags = DOCA_FLOW_NO_WAIT;
			result = doca_flow_pipe_rm_entry(0, flags, &status, (aged_entry_data->entry));
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_INFO("Failed to remove aged entry: %s", doca_get_error_string(result));
				return result;
			}
		}

		do {
			result = doca_flow_entries_process(port, 0, DEFAULT_TIMEOUT_US, num_of_aged_entries - status.nb_processed);
			if (result != DOCA_SUCCESS) {
				DOCA_LOG_ERR("Failed to process entries: %s", doca_get_error_string(result));
				return result;
			}
			if (status.failure) {
				DOCA_LOG_ERR("Failed to process entries, status is not success");
				return DOCA_ERROR_BAD_STATE;
			}
		} while (status.nb_processed < num_of_aged_entries);

		num_of_aged_entries = doca_flow_aging_handle(port, 0, quota_time, aged_entries, num_of_aging_entries);
	}
	return DOCA_SUCCESS;
}

/*
 * Run flow_aging sample
 *
 * @nb_queues [in]: number of queues the sample will use
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise.
 */
doca_error_t
flow_aging(int nb_queues)
{
	const int nb_ports = 2;
	/* the counters will divide by all queues per port */
	struct doca_flow_resources resource = {0};
	uint32_t nr_shared_resources[DOCA_FLOW_SHARED_RESOURCE_MAX] = {0};
	struct doca_flow_port *ports[nb_ports];
	struct doca_flow_pipe *pipe;
	struct doca_flow_aged_query *aged_entries;
	struct sample_user_data *user_data[nb_ports];
	int num_of_aging_entries = 10;
	int aged_entry_counter = 0;
	doca_error_t result;
	int port_id;

	resource.nb_counters = 80;

	result = init_doca_flow(nb_queues, "vnf,hws", resource, nr_shared_resources);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init DOCA Flow: %s", doca_get_error_string(result));
		return result;
	}

	result = init_doca_flow_ports(nb_ports, ports, true);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init DOCA ports: %s", doca_get_error_string(result));
		doca_flow_destroy();
		return result;
	}

	memset(user_data, 0, sizeof(struct sample_user_data *) * nb_ports);

	for (port_id = 0; port_id < nb_ports; port_id++) {
		result = create_aging_pipe(ports[port_id], port_id, &pipe);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to create pipe: %s", doca_get_error_string(result));
			goto entries_cleanup;
		}

		user_data[port_id] = (struct sample_user_data *)malloc(
			num_of_aging_entries * sizeof(struct sample_user_data));
		if (user_data[port_id] == NULL) {
			DOCA_LOG_ERR("Failed to allocate user data");
			result = DOCA_ERROR_NO_MEMORY;
			goto entries_cleanup;
		}

		result = add_aging_pipe_entries(pipe, user_data[port_id], ports[port_id], port_id, num_of_aging_entries);
		if (result != DOCA_SUCCESS) {
			stop_doca_flow_ports(nb_ports, ports);
			goto entries_cleanup;
		}
	}

	aged_entries = (struct doca_flow_aged_query *)malloc(num_of_aging_entries * nb_ports *
							     sizeof(struct sample_user_data));
	if (aged_entries == NULL) {
		DOCA_LOG_ERR("Failed to allocate aged entries array");
		result = DOCA_ERROR_NO_MEMORY;
		goto entries_cleanup;
	}
	/* handle aging in loop until all entries aged out */
	while (aged_entry_counter < num_of_aging_entries * nb_ports) {
		for (port_id = 0; port_id < nb_ports; port_id++) {
			sleep(5);
			result = handle_aged_flow(ports[port_id], aged_entries, num_of_aging_entries, &aged_entry_counter);
			if (result != DOCA_SUCCESS)
				break;
		}
	}

	free(aged_entries);
entries_cleanup:
	for (port_id = 0; port_id < nb_ports; port_id++) {
		if (user_data[port_id])
			free(user_data[port_id]);
	}
	stop_doca_flow_ports(nb_ports, ports);
	doca_flow_destroy();
	return result;
}
